MISSING_API_KEY_ERROR_MESSAGE = """No API key found for Portkey.
Please set either the PORTKEY_API_KEY environment variable or \
pass the api_key prior to initialization of Portkey.
API keys can be found or created at Portkey Dashboard \

Here's how you get it:
1. Visit https://app.portkey.ai/
1. Click on your profile icon on the top left
2. From the dropdown menu, click on "Copy API Key"
"""


DEFAULT_MAX_RETRIES = 2
VERSION = "0.1.0"
DEFAULT_TIMEOUT = 60
